/* 
 * File:   ubench.cpp
 * Author: tabrown
 *
 * An array-based microbenchmark for k-CAS
 * 
 * Created on June 23, 2016, 6:45 PM
 */

#include <cstdlib>
#include <chrono>
#include <time.h>
#include <pthread.h>
#include <math.h>
#include <atomic>
#include <algorithm>
#include <set>
#include <string>
#include <iostream>
#include "../debugcounter.h"
#include "../random.h"
#include "kcas_reuse.h"
using namespace std;

#ifndef DEBUG
#define DEBUG if(0)
#define DEBUG1 if(0)
#define DEBUG2 if(0)
#endif

#ifndef TRACE
#define TRACE if(0)
#endif

#ifndef COUTATOMICTID
#define COUTATOMICTID(str) cout<<"tid="<<tid<<": "<<str;
#endif

#ifndef COUTATOMIC
#define COUTATOMIC(str) cout<<str;
#endif

#ifndef VERBOSE
#define VERBOSE if(0)
#endif

#ifndef MEM_CONTEND_UNIT_WORDS
#define MEM_CONTEND_UNIT_WORDS 8
#endif

#ifndef MEM_CONTEND_UNIT_BYTES
#define MEM_CONTEND_UNIT_BYTES 128
#endif

int MAXKEY;
int MILLIS_TO_RUN;
int TOTAL_THREADS;

#ifndef KCAS_MAXK
#error Must define KCAS_MAXK
#endif

#ifndef KCAS_MAXTHREADS
#error Must define KCAS_MAXTHREADS
#endif

typedef kcasdesc_t<KCAS_MAXK, KCAS_MAXTHREADS> DescriptorType;
typedef rdcssdesc_t DescriptorType2;

typedef kcasProvider<KCAS_MAXK,KCAS_MAXTHREADS> KCASProviderType;

#define stringify(x) #x
#define dbg(x) " "<<stringify(x)<<"="<<x

class dataStructure {
public:
    KCASProviderType prov;
    casword_t *data;
    const int size;
    debugCounter *successful;
    debugCounter *totalOps;

    dataStructure(const int numProcesses, const int _size) : size(_size) {
        data = new casword_t[_size];
        for (int i=0;i<_size;++i) {
            prov.writeVal(&data[i], 0);
        }
        successful = new debugCounter(numProcesses);
        totalOps = new debugCounter(numProcesses);
    }
    ~dataStructure() {
        delete[] data;
        delete successful;
        delete totalOps;
    }
    bool atomicIncrementK(const int tid, Random * const rng) {
        // generate random indices to perform kcas on
        int ix[KCAS_MAXK+1];
        
        {
            int maxval = -1;
            for (int i=0;i<KCAS_MAXK+1;++i) {
                ix[i] = rng->nextNatural() & 0x7fffffff;
                if (ix[i] > maxval) maxval = ix[i];
            }
            sort(ix, ix + KCAS_MAXK+1);
            int maxover = 0;
            for (int i=0;i<KCAS_MAXK;++i) {
                ix[i] = (int)((double) ix[i] * MAXKEY / maxval);
                if (ix>0 && ix[i] <= ix[i-1]) ix[i] = ix[i-1]+1;
                if (ix[i] - (MAXKEY-1) > maxover) maxover = ix[i] - (MAXKEY-1);
            }
            if (maxover > 0) {
                //cout<<"maxover was "<<maxover<<endl;
                if (ix[0] - maxover >= 0) ix[0] -= maxover;
                else ix[0] = 0;
                
                for (int i=1;i<KCAS_MAXK;++i) {
                    if (ix[i] - maxover > ix[i-1]) ix[i] -= maxover;
                    else ix[i] = ix[i-1]+1;
                }
            }
        }
        
#ifndef NDEBUG
        // perform validation on indices generated by the above
        for (int i=0;i<KCAS_MAXK;++i) {
            if (ix[i] < 0) { cout<<"FAILURE: ix["<<i<<"]="<<ix[i]<<endl; exit(-1); }
            if (ix[i] >= MAXKEY) { cout<<"FAILURE: ix["<<i<<"]="<<ix[i]<<endl; exit(-1); }
            if (i>0 && ix[i] <= ix[i-1]) { cout<<"FAILURE: ix["<<i-1<<"]="<<ix[i-1]<<" and ix["<<i<<"]="<<ix[i]<<endl; exit(-1); }
        }
#endif
        
        // create a new kcas descriptor
        DescriptorType *ptr = prov.createKcasDesc(tid);
        for (int i=0;i<KCAS_MAXK;++i) {
            casword_t* addr = &data[ix[i]];
            casword_t oldval = prov.readVal(tid, &data[ix[i]]);
            casword_t newval = oldval+1;
            ptr->addValAddr(addr, oldval, newval);
        }
        
        // invoke the actual kcas
        return prov.kcas(tid, ptr);
    }
    long long getTotal(const int tidForReading) {
        long long result = 0;
        for (int i=0;i<size;++i) {
            result += prov.readVal(tidForReading, &data[i]);
        }
        return result;
    }
};

dataStructure *ds;

volatile char padding0[MEM_CONTEND_UNIT_BYTES];
static Random rngs[MAX_TID_POW2*MEM_CONTEND_UNIT_WORDS]; // create per-thread random number generators (padded to avoid false sharing)
volatile char padding1[MEM_CONTEND_UNIT_BYTES];

// variables used in the concurrent test
chrono::time_point<chrono::high_resolution_clock> startTime;
chrono::time_point<chrono::high_resolution_clock> endTime;
volatile char padding2[MEM_CONTEND_UNIT_BYTES];
long elapsedMillis;
volatile char padding3[MEM_CONTEND_UNIT_BYTES];
bool start = false;
volatile char padding4[MEM_CONTEND_UNIT_BYTES];
bool done = false;
volatile char padding5[MEM_CONTEND_UNIT_BYTES];
atomic_int running; // number of threads that are running
volatile char padding6[MEM_CONTEND_UNIT_BYTES];

void *thread_timed(void *_id) {
    const int OPS_BETWEEN_TIME_CHECKS = 500;
    int tid = *((int*) _id);
    Random *rng = &rngs[tid*MEM_CONTEND_UNIT_WORDS];

    running.fetch_add(1);
    __sync_synchronize();
    while (!start) { __sync_synchronize(); TRACE COUTATOMICTID("waiting to start"<<endl); } // wait to start
    int cnt = 0;
    while (!done) {
        if (((++cnt) % OPS_BETWEEN_TIME_CHECKS) == 0) {
            chrono::time_point<chrono::high_resolution_clock> __endTime = chrono::high_resolution_clock::now();
            if (chrono::duration_cast<chrono::milliseconds>(__endTime-startTime).count() >= MILLIS_TO_RUN) {
                done = true;
                __sync_synchronize();
                break;
            }
        }
        
        VERBOSE if (cnt&&((cnt % 1000000) == 0)) COUTATOMICTID("op# "<<cnt<<endl);
        if (ds->atomicIncrementK(tid, rng)) {
            ds->successful->inc(tid);
        }
        ds->totalOps->inc(tid);
    }
    running.fetch_add(-1);
//    COUTATOMICTID("terminated"<<endl);
    return NULL;
}

void trial() {
    // get random number generator seeded with time
    // we use this rng to seed per-thread rng's that use a different algorithm
    srand(time(NULL));

    // create threads
    pthread_t *threads[TOTAL_THREADS];
    int ids[TOTAL_THREADS];
    for (int i=0;i<TOTAL_THREADS;++i) {
        threads[i] = new pthread_t;
        ids[i] = i;
        rngs[i*MEM_CONTEND_UNIT_WORDS].setSeed(rand());
    }
    
    // amount of time for main thread to wait for children threads
    timespec tsExpected;
    tsExpected.tv_sec = MILLIS_TO_RUN / 1000;
    tsExpected.tv_nsec = (MILLIS_TO_RUN % 1000) * ((__syscall_slong_t) 1000000);
    // short nap
    timespec tsNap;
    tsNap.tv_sec = 0;
    tsNap.tv_nsec = 10000000; // 10ms

    // start all threads
    for (int i=0;i<TOTAL_THREADS;++i) {
        if (pthread_create(threads[i], NULL, thread_timed, &ids[i])) {
            cerr<<"ERROR: could not create thread"<<endl;
            exit(-1);
        }
    }

    while (running.load() < TOTAL_THREADS) {
        TRACE COUTATOMIC("main thread: waiting for threads to START running="<<running.load()<<endl);
    } // wait for all threads to be ready

    COUTATOMIC("main thread: starting timer..."<<endl);
    startTime = chrono::high_resolution_clock::now();
    __sync_synchronize();
    start = true;

    // replace pthread_join with sleeping, and kill if we run too long
    // method: sleep for the desired time + a small epsilon,
    //      then check "running.load()" to see if we're done.
    //      if not, loop and sleep in small increments for up to 1s,
    //      and exit(-1) if running doesn't hit 0.

    nanosleep(&tsExpected, NULL); //&tsElapsed);
    done = true;
    elapsedMillis = chrono::duration_cast<chrono::milliseconds>(chrono::high_resolution_clock::now() - startTime).count();
    long elapsedMillisNapping = 0;
    while (running.load() > 0 && elapsedMillisNapping < 200) { // nap for up to 200ms
        nanosleep(&tsNap, NULL); //&tsElapsed);
        elapsedMillisNapping = chrono::duration_cast<chrono::milliseconds>(chrono::high_resolution_clock::now() - startTime).count() - elapsedMillis;
    }
    elapsedMillis += elapsedMillisNapping;
#if 1
    if (running.load() > 0) {
        cout<<endl;
        COUTATOMIC("Validation FAILURE: "<<running.load()<<" non-terminating thread(s)"<<endl);
        cout<<endl;
        cout<<"ops="<<ds->successful->getTotal()<<endl;
        const int tidForReading = 0;
        cout<<"arrayTotal/K="<<(ds->getTotal(tidForReading)/KCAS_MAXK)<<endl;
        exit(-1);
    }
#else
    while (running.load() > 0) {}
#endif
    // join all threads
    for (int i=0;i<TOTAL_THREADS;++i) {
        if (pthread_join(*(threads[i]), NULL)) {
            cerr<<"ERROR: could not join thread"<<endl;
            exit(-1);
        }
    }

    COUTATOMIC((elapsedMillis/1000.)<<"s"<<endl);
    
    for (int i=0;i<TOTAL_THREADS;++i) {
        delete threads[i];
    }
}

#define PRINTI(name) { cout<<(#name)<<"="<<name<<endl; }
#define PRINTS(name) { cout<<(#name)<<"="<<name<<endl; }

void printOutput() {
    const int tid = 0;
    
    long long successfulOps = ds->successful->getTotal();
    long long totalOps = ds->totalOps->getTotal();
    
    long long sumOfEntries = ds->getTotal(tid);
    cout<<"TOTAL="<<sumOfEntries<<endl;
    long long scaled = sumOfEntries/KCAS_MAXK;

    cout<<"Validation: "<<successfulOps<<" (#ops) ==? "<<scaled<<" (scaled total): ";
    cout<<((successfulOps == scaled) ? "OK." : "FAILED.")<<endl;
    cout<<endl;

    //cout<<"sum of entries : "<<sumOfEntries<<endl;

    cout<<"completed ops       : "<<totalOps<<endl;
    cout<<"throughput           : "<<(long long)(totalOps * 1000. / elapsedMillis)<<endl;
    cout<<"    incl. queries    : "<<(long long)(totalOps * 1000. / elapsedMillis)<<endl;
    cout<<"elapsed milliseconds : "<<elapsedMillis<<endl;
    cout<<endl;
    
    if (successfulOps != scaled) {
        cout<<"ERROR: validation failed."<<endl;
        exit(-1);
    }
    
}

int main(int argc, char** argv) {
    MILLIS_TO_RUN = -1;
    for (int i=1;i<argc;++i) {
        if (strcmp(argv[i], "-k") == 0) {
            MAXKEY = atoi(argv[++i]);
        } else if (strcmp(argv[i], "-n") == 0) {
            TOTAL_THREADS = atoi(argv[++i]);
        } else if (strcmp(argv[i], "-t") == 0) {
            MILLIS_TO_RUN = atoi(argv[++i]);
        } else {
            cout<<"bad arguments"<<endl;
            exit(1);
        }
    }
    
    PRINTI(KCAS_MAXK);
    PRINTI(KCAS_MAXTHREADS);
    PRINTI(MILLIS_TO_RUN);
    PRINTI(MAXKEY);
    PRINTI(TOTAL_THREADS);
#ifdef WIDTH_SEQ
    PRINTI(WIDTH_SEQ);
#endif
    
    ds = new dataStructure(TOTAL_THREADS, MAXKEY);

    trial();
    printOutput();
    
    delete ds;
    
    return 0;
}
